#!/bin/bash
# This script is meant to be called by the "after_success" step defined in
# .travis.yml. See http://docs.travis-ci.com/ for more details.

# License: 3-clause BSD

set -e

# Need to run covdecov from a git checkout, so we copy .coverage
# from TEST_DIR where pytest has been run
cd $TRAVIS_BUILD_DIR
# Ignore covdecov failures as the covdecov server is not
# very reliable but we don't want travis to report a failure
# in the github UI just because the coverage report failed to
# be published.
codecov || echo "Covdecov upload failed"
